@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 #
 # Module providing the `Pool` class for managing a process pool
 #
@@ -7,7 +6,6 @@
 # Copyright (c) 2006-2008, R Oudkerk
 # Licensed to PSF under a Contributor Agreement.
 #
-from __future__ import absolute_import
 
 #
 # Imports
@@ -43,8 +41,9 @@ from .exceptions import (
     TimeoutError,
     WorkerLostError,
 )
-from .five import Empty, Queue, range, values, reraise, monotonic
-from .util import Finalize, debug
+from time import monotonic
+from queue import Queue, Empty
+from .util import Finalize, debug, warning
 
 MAXMEM_USED_FMT = """\
 child process exiting after exceeding memory limit ({0}KiB / {1}KiB)
@@ -110,6 +109,8 @@ SIG_SOFT_TIMEOUT = getattr(signal, "SIGUSR1", None)
 
 LOST_WORKER_TIMEOUT = 10.0
 EX_OK = getattr(os, "EX_OK", 0)
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT = 300
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL = 0.1
 
 job_counter = itertools.count()
 
@@ -209,7 +210,7 @@ class MaybeEncodingError(Exception):
     def __init__(self, exc, value):
         self.exc = repr(exc)
         self.value = repr(value)
-        super(MaybeEncodingError, self).__init__(self.exc, self.value)
+        super().__init__(self.exc, self.value)
 
     def __repr__(self):
         return "<%s: %s>" % (self.__class__.__name__, str(self))
@@ -231,12 +232,12 @@ def soft_timeout_sighandler(signum, frame):
 #
 
 
-class Worker(object):
+class Worker:
 
     def __init__(self, inq, outq, synq=None, initializer=None, initargs=(),
                  maxtasks=None, sentinel=None, on_exit=None,
                  sigprotection=True, wrap_exception=True,
-                 max_memory_per_child=None):
+                 max_memory_per_child=None, on_ready_counter=None):
         assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
         self.initializer = initializer
         self.initargs = initargs
@@ -247,6 +248,7 @@ class Worker(object):
         self.sigprotection = sigprotection
         self.inq, self.outq, self.synq = inq, outq, synq
         self.wrap_exception = wrap_exception  # XXX cannot disable yet
+        self.on_ready_counter = on_ready_counter
         self.contribute_to_object(self)
 
     def contribute_to_object(self, obj):
@@ -268,7 +270,7 @@ class Worker(object):
         return self.__class__, (
             self.inq, self.outq, self.synq, self.initializer,
             self.initargs, self.maxtasks, self._shutdown, self.on_exit,
-            self.sigprotection, self.wrap_exception,
+            self.sigprotection, self.wrap_exception, self.max_memory_per_child,
         )
 
     def __call__(self):
@@ -277,7 +279,7 @@ class Worker(object):
 
         def exit(status=None):
             _exitcode[0] = status
-            return _exit()
+            return _exit(status)
         sys.exit = exit
 
         pid = os.getpid()
@@ -343,47 +345,70 @@ class Worker(object):
                 i += 1
 
         completed = 0
-        while maxtasks is None or (maxtasks and completed < maxtasks):
-            req = wait_for_job()
-            if req:
-                type_, args_ = req
-                assert type_ == TASK
-                job, i, fun, args, kwargs = args_
-                put((ACK, (job, i, now(), pid, synqW_fd)))
-                if _wait_for_syn:
-                    confirm = wait_for_syn(job)
-                    if not confirm:
-                        continue  # received NACK
-                try:
-                    result = (True, prepare_result(fun(*args, **kwargs)))
-                except Exception:
-                    result = (False, ExceptionInfo())
-                try:
-                    put((READY, (job, i, result, inqW_fd)))
-                except Exception as exc:
-                    _, _, tb = sys.exc_info()
+        try:
+            while maxtasks is None or (maxtasks and completed < maxtasks):
+                req = wait_for_job()
+                if req:
+                    type_, args_ = req
+                    assert type_ == TASK
+                    job, i, fun, args, kwargs = args_
+                    put((ACK, (job, i, now(), pid, synqW_fd)))
+                    if _wait_for_syn:
+                        confirm = wait_for_syn(job)
+                        if not confirm:
+                            continue  # received NACK
                     try:
-                        wrapped = MaybeEncodingError(exc, result[1])
-                        einfo = ExceptionInfo((
-                            MaybeEncodingError, wrapped, tb,
-                        ))
-                        put((READY, (job, i, (False, einfo), inqW_fd)))
-                    finally:
-                        del(tb)
-                completed += 1
-                if max_memory_per_child > 0:
-                    used_kb = mem_rss()
-                    if used_kb <= 0:
-                        error('worker unable to determine memory usage')
-                    if used_kb > 0 and used_kb > max_memory_per_child:
-                        error(MAXMEM_USED_FMT.format(
-                            used_kb, max_memory_per_child))
-                        return EX_RECYCLE
+                        result = (True, prepare_result(fun(*args, **kwargs)))
+                    except Exception:
+                        result = (False, ExceptionInfo())
+                    try:
+                        put((READY, (job, i, result, inqW_fd)))
+                    except Exception as exc:
+                        _, _, tb = sys.exc_info()
+                        try:
+                            wrapped = MaybeEncodingError(exc, result[1])
+                            einfo = ExceptionInfo((
+                                MaybeEncodingError, wrapped, tb,
+                            ))
+                            put((READY, (job, i, (False, einfo), inqW_fd)))
+                        finally:
+                            del(tb)
+                    completed += 1
+                    if max_memory_per_child > 0:
+                        used_kb = mem_rss()
+                        if used_kb <= 0:
+                            error('worker unable to determine memory usage')
+                        if used_kb > 0 and used_kb > max_memory_per_child:
+                            warning(MAXMEM_USED_FMT.format(
+                                used_kb, max_memory_per_child))
+                            return EX_RECYCLE
 
-        debug('worker exiting after %d tasks', completed)
-        if maxtasks:
-            return EX_RECYCLE if completed == maxtasks else EX_FAILURE
-        return EX_OK
+            debug('worker exiting after %d tasks', completed)
+            if maxtasks:
+                return EX_RECYCLE if completed == maxtasks else EX_FAILURE
+            return EX_OK
+        finally:
+            # Before exiting the worker, we want to ensure that that all
+            # messages produced by the worker have been consumed by the main
+            # process. This prevents the worker being terminated prematurely
+            # and messages being lost.
+            self._ensure_messages_consumed(completed=completed)
+
+    def _ensure_messages_consumed(self, completed):
+        """ Returns true if all messages sent out have been received and
+        consumed within a reasonable amount of time """
+
+        if not self.on_ready_counter:
+            return False
+
+        for retry in range(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT):
+            if self.on_ready_counter.value >= completed:
+                debug('ensured messages consumed after %d retries', retry)
+                return True
+            time.sleep(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL)
+        warning('could not ensure all messages were consumed prior to '
+                'exiting')
+        return False
 
     def after_fork(self):
         if hasattr(self.inq, '_writer'):
@@ -510,7 +535,7 @@ class Supervisor(PoolThread):
 
     def __init__(self, pool):
         self.pool = pool
-        super(Supervisor, self).__init__()
+        super().__init__()
 
     def body(self):
         debug('worker handler starting')
@@ -530,8 +555,8 @@ class Supervisor(PoolThread):
                     pool._maintain_pool()
                     time.sleep(0.1)
 
-            # Keep maintaing workers until the cache gets drained, unless
-            # the pool is termianted
+            # Keep maintaining workers until the cache gets drained, unless
+            # the pool is terminated
             pool.restart_state = prev_state
             while self._state == RUN and pool._state == RUN:
                 pool._maintain_pool()
@@ -551,7 +576,7 @@ class TaskHandler(PoolThread):
         self.outqueue = outqueue
         self.pool = pool
         self.cache = cache
-        super(TaskHandler, self).__init__()
+        super().__init__()
 
     def body(self):
         cache = self.cache
@@ -626,7 +651,7 @@ class TimeoutHandler(PoolThread):
         self.t_soft = t_soft
         self.t_hard = t_hard
         self._it = None
-        super(TimeoutHandler, self).__init__()
+        super().__init__()
 
     def _process_by_pid(self, pid):
         return next((
@@ -694,7 +719,6 @@ class TimeoutHandler(PoolThread):
             pass
 
     def handle_timeouts(self):
-        cache = copy.deepcopy(self.cache)
         t_hard, t_soft = self.t_hard, self.t_soft
         dirty = set()
         on_soft_timeout = self.on_soft_timeout
@@ -708,12 +732,16 @@ class TimeoutHandler(PoolThread):
 
         # Inner-loop
         while self._state == RUN:
+            # Perform a shallow copy before iteration because keys can change.
+            # A deep copy fails (on shutdown) due to thread.lock objects.
+            # https://github.com/celery/billiard/issues/260
+            cache = copy.copy(self.cache)
 
             # Remove dirty items not in cache anymore
             if dirty:
                 dirty = set(k for k in dirty if k in cache)
 
-            for i, job in list(cache.items()):
+            for i, job in cache.items():
                 ack_time = job._time_accepted
                 soft_timeout = job._soft_timeout
                 if soft_timeout is None:
@@ -750,7 +778,7 @@ class ResultHandler(PoolThread):
 
     def __init__(self, outqueue, get, cache, poll,
                  join_exited_workers, putlock, restart_state,
-                 check_timeouts, on_job_ready):
+                 check_timeouts, on_job_ready, on_ready_counters=None):
         self.outqueue = outqueue
         self.get = get
         self.cache = cache
@@ -762,8 +790,9 @@ class ResultHandler(PoolThread):
         self._shutdown_complete = False
         self.check_timeouts = check_timeouts
         self.on_job_ready = on_job_ready
+        self.on_ready_counters = on_ready_counters
         self._make_methods()
-        super(ResultHandler, self).__init__()
+        super().__init__()
 
     def on_stop_not_started(self):
         # used when pool started without result handler thread.
@@ -790,6 +819,14 @@ class ResultHandler(PoolThread):
                 item = cache[job]
             except KeyError:
                 return
+
+            if self.on_ready_counters:
+                worker_pid = next(iter(item.worker_pids()), None)
+                if worker_pid and worker_pid in self.on_ready_counters:
+                    on_ready_counter = self.on_ready_counters[worker_pid]
+                    with on_ready_counter.get_lock():
+                        on_ready_counter.value += 1
+
             if not item.ready():
                 if putlock is not None:
                     putlock.release()
@@ -922,7 +959,7 @@ class ResultHandler(PoolThread):
               len(cache), self._state)
 
 
-class Pool(object):
+class Pool:
     '''
     Class which supports an async version of applying functions to arguments.
     '''
@@ -1001,6 +1038,7 @@ class Pool(object):
 
         self._pool = []
         self._poolctrl = {}
+        self._on_ready_counters = {}
         self.putlocks = putlocks
         self._putlock = semaphore or LaxBoundedSemaphore(self._processes)
         for i in range(self._processes):
@@ -1066,7 +1104,8 @@ class Pool(object):
             self._outqueue, self._quick_get, self._cache,
             self._poll_result, self._join_exited_workers,
             self._putlock, self.restart_state, self.check_timeouts,
-            self.on_job_ready, **extra_kwargs
+            self.on_job_ready, on_ready_counters=self._on_ready_counters,
+            **extra_kwargs
         )
 
     def on_job_ready(self, job, i, obj, inqW_fd):
@@ -1099,6 +1138,7 @@ class Pool(object):
     def _create_worker_process(self, i):
         sentinel = self._ctx.Event() if self.allow_restart else None
         inq, outq, synq = self.get_process_queues()
+        on_ready_counter = self._ctx.Value('i')
         w = self.WorkerProcess(self.Worker(
             inq, outq, synq, self._initializer, self._initargs,
             self._maxtasksperchild, sentinel, self._on_process_exit,
@@ -1107,6 +1147,7 @@ class Pool(object):
             sigprotection=self.threads,
             wrap_exception=self._wrap_exception,
             max_memory_per_child=self._max_memory_per_child,
+            on_ready_counter=on_ready_counter,
         ))
         self._pool.append(w)
         self._process_register_queues(w, (inq, outq, synq))
@@ -1115,6 +1156,7 @@ class Pool(object):
         w.index = i
         w.start()
         self._poolctrl[w.pid] = sentinel
+        self._on_ready_counters[w.pid] = on_ready_counter
         if self.on_process_up:
             self.on_process_up(w)
         return w
@@ -1165,6 +1207,7 @@ class Pool(object):
                 self.process_flush_queues(worker)
                 del self._pool[i]
                 del self._poolctrl[worker.pid]
+                del self._on_ready_counters[worker.pid]
         if cleaned:
             all_pids = [w.pid for w in self._pool]
             for job in list(self._cache.values()):
@@ -1196,7 +1239,7 @@ class Pool(object):
                     elif sched_for and not sched_for._is_alive():
                         self.on_job_process_down(job, sched_for.pid)
 
-            for worker in values(cleaned):
+            for worker in cleaned.values():
                 if self.on_process_down:
                     if not shutdown:
                         self._process_cleanup_queues(worker)
@@ -1219,8 +1262,8 @@ class Pool(object):
     def mark_as_worker_lost(self, job, exitcode):
         try:
             raise WorkerLostError(
-                'Worker exited prematurely: {0}.'.format(
-                    human_status(exitcode)),
+                'Worker exited prematurely: {0} Job: {1}.'.format(
+                    human_status(exitcode), job._job),
             )
         except WorkerLostError:
             job._set(None, (False, ExceptionInfo()))
@@ -1311,9 +1354,7 @@ class Pool(object):
                 raise
             except OSError as exc:
                 if get_errno(exc) == errno.ENOMEM:
-                    reraise(MemoryError,
-                            MemoryError(str(exc)),
-                            sys.exc_info()[2])
+                    raise MemoryError from exc
                 raise
 
     def _setup_queues(self):
@@ -1650,6 +1691,11 @@ class Pool(object):
                         p.join()
             debug('pool workers joined')
 
+        if inqueue:
+            inqueue.close()
+        if outqueue:
+            outqueue.close()
+
     @property
     def process_sentinels(self):
         return [w._popen.sentinel for w in self._pool]
@@ -1659,7 +1705,7 @@ class Pool(object):
 #
 
 
-class ApplyResult(object):
+class ApplyResult:
     _worker_lost = None
     _write_to = None
     _scheduled_for = None
@@ -1701,7 +1747,7 @@ class ApplyResult(object):
         )
 
     def ready(self):
-        return self._event.isSet()
+        return self._event.is_set()
 
     def accepted(self):
         return self._accepted
@@ -1878,7 +1924,7 @@ class MapResult(ApplyResult):
 #
 
 
-class IMapIterator(object):
+class IMapIterator:
     _worker_lost = None
 
     def __init__(self, cache, lost_worker_timeout=LOST_WORKER_TIMEOUT):
